;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "telega" "20260706.2147"
  "Telegram client (unofficial)."
  '((emacs              "27.1")
    (visual-fill-column "1.9")
    (transient          "0.9.0"))
  :url "https://github.com/zevlg/telega.el"
  :commit "fb5571bbb49304b0f61b208edde396c773f6285a"
  :revdesc "fb5571bbb493"
  :keywords '("comm")
  :authors '(("Zajcev Evgeny" . "zevlg@yandex.ru"))
  :maintainers '(("Zajcev Evgeny" . "zevlg@yandex.ru")))
